/*
patchlev.h
copyright 2009-2024,2025, Thomas E. Dickey
copyright 1991-1996,2014, Michael D. Brennan

This is a source file for mawk, an implementation of
the AWK programming language.

Mawk is distributed without warranty under the terms of
the GNU General Public License, version 2, 1991.
 */

/*
 * $MawkId: patchlev.h,v 1.168 2025/01/31 22:21:11 tom Exp $
 */
#define  PATCH_BASE	1
#define  PATCH_LEVEL	3
#define  PATCH_STRING	".4"
#define  DATE_STRING    "20250131"
